# -*- coding: utf-8 -*-
"""
Created on Thu May 30 13:45:21 2024

@author: dvezinet
"""

import copy


# ##################################################################
# ##################################################################
#                   Default dict
# ##################################################################


_DAVAIL = {

    'standard - LOS': {
        'conditions': None,
        'fields': {
            'etendue': {
                'doc': 'etendue per pixel, from LOS',
            },
            # 'amin': {
            #     'doc': '',
            # },
            # 'amax': {
            #     'doc': '',
            # },
            'length': {
                'doc': 'length of the LOS',
            },
            'tangency_radius': {
                'doc': 'minimum major radius of LOS',
            },
            'alpha': {
                'doc': 'incidence angle of LOS on PFC (end point)',
            },
            'alpha_pixel': {
                'doc': 'incidence angle of LOS on pixel (start point)',
            },
        },
    },

    'broadband - VOS': {
        'conditions': {'spectro': False, 'is_vos': True},
        'fields': {
            'vos_pix_sang_integ': {
                'doc': 'integral of all solid angle from pts, per pixel',
            },
            # 'vos_cross_ndet': {
            #     'doc': 'sum of all solid angle from pts, per pixel',
            # },
            # 'vos_cross_sang': {
            #     'doc': 'sum of all solid angle to pixels, per pts',
            # },
            # 'vos_cross_ang_pol_target': {
            #     'doc': 'sum of all solid angle to pixels, per pts',
            # },
            # 'vos_cross_ang_pol_score': {
            #     'doc': 'sum of all solid angle to pixels, per pts',
            # },
            # 'vos_cross_ang_tor_target': {
            #     'doc': 'sum of all solid angle to pixels, per pts',
            # },
            # 'vos_cross_ang_tor_score': {
            #     'doc': 'sum of all solid angle to pixels, per pts',
            # },
        },
    },

    'spectro - VOS': {
        'conditions': {'spectro': True, 'is_vos': True},
        'fields': {
            'vos_lamb': {
                'doc': 'reference wavelength, from VOS, per pixel',
            },
            'vos_dlamb': {
                'doc': '(lambmax - lambmin), from VOS, per pixel',
            },
            'vos_ph_integ': {
                'doc': 'sum of all ph over pts and wavelengths, per pixel',
            },
        },
    },

    'spectro - lamb': {
        'conditions': {'spectro': True},
        'fields': {
            'lamb': {
                'doc': 'reference wavelength, from LOS, per pixel',
            },
            'lambmin': {
                'doc': 'minimal wavelgenth, from LOS, per pixel',
            },
            'lambmax': {
                'doc': 'minimal wavelgenth, from LOS, per pixel',
            },
            'dlamb': {
                'doc': '(lambmax - lambmin), from LOS, per pixel',
            },
            'res': {
                'doc': 'lamb / (lambmax - lambmin), from LOS, per pixel',
            },
        },
    },

    # '': {
    #     'conditions': None,
    #     'fields': {
    #         '': {
    #             'doc': '',
    #         },
    #     },
    # },
}


# ##################################################################
# ##################################################################
#                   function
# ##################################################################


def get_davail(
    coll=None,
    key=None,
    key_cam=None,
    # conditions
    **kwdargs, # spectro, is_vos, is_3d
):

    # --------------
    # initialize
    # --------------

    davail = copy.deepcopy(_DAVAIL)

    # lcomp += llamb

    # -------------------------
    # update for specific diag
    # -------------------------

    lok = []
    if key is not None:

        # --------------------
        # reduce to applicable

        for k0, v0 in davail.items():

            c0 = (
                v0['conditions'] is None
                or all([
                    kwdargs.get(cc) is vv
                    for cc, vv in v0['conditions'].items()
                ])
            )

            if c0:
                lok.append(k0)

        # update
        davail = {k0: davail[k0] for k0 in lok}

        # ----------------
        # add parameters


        # -------------------------------------
        # add raw data (fixed pixel-wise data directly from ddata)

        if len(key_cam) == 1:
            davail['raw'] = {
                'fields': {
                    k0: {'doc': 'raw static data (1 camera)'}
                    for k0, v0 in coll.ddata.items()
                    if v0['ref'] == coll.dobj['camera'][key_cam[0]]['dgeom']['ref']
                },
            }

        # ------------------
        # add synthetic data

        lsynth = coll.dobj['diagnostic'][key]['signal']
        if lsynth is not None and len(lsynth) > 0:
            davail['synth'] = {
                'fields': {
                    k0: {'doc': 'synthetic data'}
                    for k0 in lsynth
                },
            }

    return davail




import numpy as np
import scipy.interpolate as scpinterp
import scipy.stats as scpstats
import Polygon as plg
from matplotlib.path import Path
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from mpl_toolkits.mplot3d import Axes3D
import datastock as ds


from . import _class8_equivalent_apertures as _equivalent_apertures
from . import _class8_plot
from . import _generic_check
from . import _class8_vos_utilities as _utilities


# ################################################
# ################################################
#               main
# ################################################


def _from_pts(
    coll=None,
    # diag
    key=None,
    key_cam=None,
    # mesh sampling
    key_mesh=None,
    res_RZ=None,
    res_phi=None,
    # pts coordinates
    ptsx=None,
    ptsy=None,
    ptsz=None,
    # res
    n0=None,
    n1=None,
    min_threshold=None,
    # optional lamb
    lamb0=None,
    res_lamb=None,
    rocking_curve=None,
    res_rock_curve=None,
    # options
    append=None,
    plot=None,
    plot_pixels=None,
    plot_config=None,
    plot_rays=None,
    vmin=None,
    vmax=None,
    aspect3d=None,
    elements=None,
    colorbar=None,
    # options
    dax=None,
    fs=None,
    dmargin=None,
    wintit=None,
):

    # -------------
    # check inputs

    (
        key, key_cam, key_mesh,
        dsamp, res_phi,
        shape, iok,
        ptsx, ptsy, ptsz, phi,
        lamb0, rocking_curve,
        append, plot, plot_pixels, colorbar,
    ) = _check(
        coll=coll,
        # diag
        key=key,
        key_cam=key_cam,
        # mesh sampling
        key_mesh=key_mesh,
        res_RZ=res_RZ,
        res_phi=res_phi,
        # pts coordinates
        ptsx=ptsx,
        ptsy=ptsy,
        ptsz=ptsz,
        # optional lamb
        lamb0=lamb0,
        rocking_curve=rocking_curve,
        # options
        append=append,
        plot=plot,
        plot_pixels=plot_pixels,
        colorbar=colorbar,
    )

    # ----------------------
    # prepare optics, camera

    (
        kspectro,
        lpoly_post,
        p0x, p0y, p0z,
        nin, e0, e1,
        ptsvect_plane,
        ptsvect_spectro,
        ptsvect_cam,
        coords_x01toxyz_plane,
        cent,
        cent_cam,
        dist_to_cam,
        pix_size,
        cbin0, cbin1,
    ) = _prepare_optics(
        coll=coll,
        key=key,
        key_cam=key_cam,
    )
    shape_cam = coll.dobj['camera'][key_cam]['dgeom']['shape']

    # -------------------------------------
    # prepare lambda, angles, rocking_curve

    (
        nlamb,
        lamb,
        dlamb,
        pow_interp,
        ang_rel,
        dang,
        bragg,
    ) = _prepare_lamb(
        coll=coll,
        key_diag=key,
        key_cam=key_cam,
        kspectro=kspectro,
        lamb=lamb0,
        res_lamb=res_lamb,
        rocking_curve=rocking_curve,
        res_rock_curve=res_rock_curve,
        verb=False,
    )

    # -------------------
    # wavelength specific

    if lamb is not None:

        # -------
        # prepare lamb per pixel

        lamb_per_pix = coll.get_diagnostic_data(
            key,
            key_cam=key_cam,
            data='lamb',
        )[0][key_cam]

        bragg_per_pix = coll.get_crystal_bragglamb(
            key=kspectro,
            lamb=lamb_per_pix,
        )[0]

    # --------------
    # prepare output

    if key_mesh is None:
        func = _loop0
    else:
        func = _loop1

    # -------------
    # compute

    dout = func(**locals())
    dout['rocking_curve'] = rocking_curve

    # -------
    # reshape

    if len(shape) > 1:
        shape_new = tuple(np.r_[shape_cam, shape])
        for k0 in ['sang']:
            dout[k0]['data'] = dout[k0]['data'].reshape(shape_new)
        for k0 in ['sang0']:
            dout[k0]['data'] = dout[k0]['data'].reshape(shape_new)

    # -------
    # plot

    if plot is True:
        _plot(
            coll=coll,
            key=key,
            key_cam=key_cam,
            dout=dout,
            cbin0=cbin0,
            cbin1=cbin1,
            # options
            plot_config=plot_config,
            plot_pixels=plot_pixels,
            vmin=vmin,
            vmax=vmax,
            aspect3d=aspect3d,
            elements=elements,
            colorbar=colorbar,
            # options
            dax=dax,
            fs=fs,
            dmargin=dmargin,
            wintit=wintit,
        )

    return dout


# ################################################
# ################################################
#               check
# ################################################


def _check(
    coll=None,
    # diag
    key=None,
    key_cam=None,
    # mesh sampling
    key_mesh=None,
    res_RZ=None,
    res_phi=None,
    # pts coordinates
    ptsx=None,
    ptsy=None,
    ptsz=None,
    # optional lamb
    lamb0=None,
    rocking_curve=None,
    # bool
    append=None,
    plot=None,
    plot_pixels=None,
    colorbar=None,
):

    # --------
    # key

    # key
    lok = [
        k0 for k0, v0 in coll._dobj.get('diagnostic', {}).items()
        if v0['spectro'] is True
    ]
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # key_cam
    lok = list(coll.dobj['diagnostic'][key]['doptics'].keys())
    key_cam = ds._generic_check._check_var(
        key_cam, 'key_cam',
        types=str,
        allowed=lok,
    )

    # ----------------------------
    # coordinates vs mesh sampling

    lc = [
        key_mesh is not None,
        ptsx is not None
    ]
    if np.sum(lc) != 1:
        msg = "Please provide key_mesh xor (ptsx, ptsy, ptsz)\n"
        raise Exception(msg)

    if lc[0]:

        lok = list(coll.dobj.get('mesh', {}).keys())
        key_mesh = ds._generic_check._check_var(
            key_mesh, 'key_mesh',
            types=str,
            allowed=lok,
        )

        dsamp = coll.get_sample_mesh(
            key=key_mesh,
            res=res_RZ,
            mode='abs',
            grid=True,
            in_mesh=True,
            # non-used
            x0=None,
            x1=None,
            Dx0=None,
            Dx1=None,
            imshow=False,
            store=False,
            kx0=None,
            kx1=None,
        )

        # res_RZ
        if res_RZ is None:
            res_RZ = 0.01

        # res_phi
        if res_phi is None:
            res_phi = 0.01

        iok, shape, phi = None, None, None

    else:

        ptsx = np.atleast_1d(ptsx)
        ptsy = np.atleast_1d(ptsy)
        ptsz = np.atleast_1d(ptsz)

        # check shapes
        dshape = {
            'ptsx': ptsx.shape,
            'ptsy': ptsy.shape,
            'ptsz': ptsz.shape,
        }
        maxsize = np.max([np.prod(v0) for v0 in dshape.values()])
        shape = [v0 for v0 in dshape.values() if np.prod(v0) == maxsize][0]

        lout = [
            k0 for k0, v0 in dshape.items()
            if v0 not in [(1,), shape]
        ]
        if len(lout) > 0:
            lstr = [f"\t-{k0}.shape: {v0}" for k0, v0 in dshape.items()]
            msg = (
                "All coordinates arrays must be of same shape:\n"
                + "\n".join(lstr)
            )
            raise Exception(msg)

        # uniformize shapes
        if shape != (1,):
            if ptsx.shape == (1,):
                ptsx = np.full(shape, ptsx[0])
            if ptsy.shape == (1,):
                ptsy = np.full(shape, ptsy[0])
            if ptsz.shape == (1,):
                ptsz = np.full(shape, ptsz[0])

        dsamp = None

    # -------
    # lamb

    if lamb0 is not None:
        lamb0 = np.atleast_1d(lamb0)

        if lc[0]:
            if lamb0.shape != (1,):
                msg = "With mesh sampling, lamb0 must be a scalar!"
                raise Exception(msg)

        else:
            if lamb0.shape == (1,):
                lamb0 = np.full(ptsx.shape, lamb0[0])

            elif lamb0.shape != ptsx.shape:
                msg = (
                    "Arg lamb0 must be the same shape as coordinates\n"
                    f"\t- ptsx.shape = {ptsx.shape}\n"
                    f"\t- lamb0.shape = {lamb0.shape}\n"
                )
                raise Exception(msg)

    # --------------
    # rocking_curve

    rocking_curve = ds._generic_check._check_var(
        rocking_curve, 'rocking_curve',
        default=True,
        types=bool,
    )

    if lamb0 is None:
        rocking_curve = False

    # -----------------------
    # store shape and flatten

    if lc[0]:
        pass

    else:
        if len(shape) > 1:
            ptsx = ptsx.ravel()
            ptsy = ptsy.ravel()
            ptsz = ptsz.ravel()
            if lamb0 is not None:
                lamb0 = lamb0.ravel()

        iok = np.isfinite(ptsx) & np.isfinite(ptsy) & np.isfinite(ptsz)
        phi = np.arctan2(ptsy, ptsx)

    # ---------
    # options

    # plot
    plot = ds._generic_check._check_var(
        plot, 'plot',
        types=bool,
        default=True,
    )

    # plot_pixel
    plot_pixels = ds._generic_check._check_var(
        plot_pixels, 'plot_pixels',
        types=bool,
        default=False,
    )

    # append
    append = ds._generic_check._check_var(
        append, 'append',
        types=bool,
        default=lc[1] and ptsx.size < 100,  #  and plot is True,
    )

    # colorbar
    colorbar = ds._generic_check._check_var(
        colorbar, 'colorbar',
        types=bool,
        default=False,
    )

    return  (
        key, key_cam, key_mesh,
        dsamp, res_phi,
        shape, iok,
        ptsx, ptsy, ptsz, phi,
        lamb0, rocking_curve,
        append, plot, plot_pixels, colorbar,
    )


# ################################################
# ################################################
#               prepare optics
# ################################################


def _prepare_optics(
    coll=None,
    key=None,
    key_cam=None,
):

    # -----------------------
    # get optics and classes
    # -----------------------

    doptics = coll.dobj['diagnostic'][key]['doptics']
    lop = doptics[key_cam]['optics']
    lop, lcls = coll.get_optics_cls(lop)

    # ------------------
    # safety check
    # ------------------

    cls_spectro = 'crystal'
    kspectro = lop[lcls.index(cls_spectro)]
    ispectro = lop.index(kspectro)
    if len(lop[:ispectro]) > 1:
        msg = "Not yet implemented optics between crystal and camera!"
        raise NotImplementedError(msg)

    # ------------------
    # list of polygons post crystal
    # ------------------

    # lpoly_post = []
    lpoly_post = [
        coll.get_optics_poly(
            key=k0,
            add_points=None,
            return_outline=False,
        )
        for k0 in lop[ispectro+1:]
    ]

    # get initial polygon of crystal
    p0x, p0y, p0z = coll.get_optics_poly(key=kspectro, add_points=None)

    # ------------------
    # unit vectors of crystal
    # ------------------

    nin = coll.dobj[cls_spectro][kspectro]['dgeom']['nin']
    e0 = coll.dobj[cls_spectro][kspectro]['dgeom']['e0']
    e1 = coll.dobj[cls_spectro][kspectro]['dgeom']['e1']

    # ------------------
    # get functions for crystal and camera
    # ------------------

    ptsvect_plane = coll.get_optics_reflect_ptsvect(key=kspectro, asplane=True)
    ptsvect_spectro = coll.get_optics_reflect_ptsvect(key=kspectro, isnorm=True)
    ptsvect_cam = coll.get_optics_reflect_ptsvect(key=key_cam, fast=True)

    coords_x01toxyz_plane = coll.get_optics_x01toxyz(
        key=kspectro,
        asplane=True,
    )

    # ------------------
    # Get centers of crystal and camera to estimate distance
    # ------------------

    cent_spectro = coll.dobj[cls_spectro][kspectro]['dgeom']['cent']
    cent_cam = coll.dobj['camera'][key_cam]['dgeom']['cent']
    dist_to_cam = np.linalg.norm(cent_spectro - cent_cam)
    pix_size = np.sqrt(coll.dobj['camera'][key_cam]['dgeom']['pix_area'])

    # ------------------
    # prepare camera bin edges
    # ------------------

    kcc = coll.dobj['camera'][key_cam]['dgeom']['cents']
    cc0 = coll.ddata[kcc[0]]['data']
    cc1 = coll.ddata[kcc[1]]['data']
    cout0, cout1 = coll.get_optics_outline(key_cam, total=False)
    cbin0 = np.r_[cc0 + np.min(cout0), cc0[-1] + np.max(cout0)]
    cbin1 = np.r_[cc1 + np.min(cout1), cc1[-1] + np.max(cout1)]

    return (
        kspectro,
        lpoly_post,
        p0x, p0y, p0z,
        nin, e0, e1,
        ptsvect_plane,
        ptsvect_spectro,
        ptsvect_cam,
        coords_x01toxyz_plane,
        cent_spectro,
        cent_cam,
        dist_to_cam,
        pix_size,
        cbin0, cbin1,
    )


# ################################################
# ################################################
#           Prepare lambda
# ################################################


def _prepare_lamb(
    coll=None,
    key_diag=None,
    key_cam=None,
    kspectro=None,
    lamb=None,
    res_lamb=None,
    rocking_curve=None,
    rocking_curve_step_width=None,
    res_rock_curve=None,
    verb=None,
):

    # ------------------
    # get lamb
    # ------------------

    if res_lamb is not None:
        lamb = coll.get_diagnostic_data(
            key=key_diag,
            key_cam=key_cam,
            data='lamb',
        )[0][key_cam]

        lambmin = np.nanmin(lamb)
        lambmax = np.nanmax(lamb)
        Dlamb = (lambmax - lambmin) * 1.1
        nlamb = int(np.ceil(Dlamb / res_lamb)) + 2
        lamb = np.linspace(lambmin - 0.2*Dlamb, lambmax + 0.2*Dlamb, nlamb)
        dlamb = lamb[1] - lamb[0]

        bragg = coll.get_crystal_bragglamb(
            key=kspectro,
            lamb=lamb,
            rocking_curve=False,
        )[0]

    elif lamb is not None:

        lamb = ds._generic_check._check_flat1darray(
            lamb, 'lamb',
            dtype=float,
            sign='>0',
            unique=True,
        )
        nlamb = lamb.size
        dlamb = None
        bragg = coll.get_crystal_bragglamb(
            key=kspectro,
            lamb=lamb,
            rocking_curve=False,
        )[0]

    # ----------------
    # get power ratio
    # ----------------

    # angle relative
    cls_spectro = 'crystal'
    kang_rel = coll.dobj[cls_spectro][kspectro]['dmat']['drock']['angle_rel']
    ang_rel = coll.ddata[kang_rel]['data']

    # power ratio
    if rocking_curve is False:
        if rocking_curve_step_width is None:
            rocking_curve_step_width = np.mean(np.diff(ang_rel))

        ang_rel = 0.5 * rocking_curve_step_width * np.r_[-1, 1]
        pow_ratio = np.r_[1, 1]

    else:
        kpow = coll.dobj[cls_spectro][kspectro]['dmat']['drock']['power_ratio']
        pow_ratio = coll.ddata[kpow]['data']

    # --------------
    # interpolation
    # ---------------

    pow_interp = scpinterp.interp1d(
        ang_rel,
        pow_ratio,
        kind='linear',
        bounds_error=False,
        fill_value=0.,
    )

    dang = np.mean(np.diff(ang_rel))

    # --------------------------------------
    # overall bragg angle with rocking curve
    # --------------------------------------

    if res_lamb is None and lamb is None:
        nlamb, lamb, dlamb, bragg = None, None, None, None

    else:

        # ------------
        # safety check

        FW = coll.dobj[cls_spectro][kspectro]['dmat']['drock']['FW']
        dd0 = bragg[0] + np.r_[0, ang_rel[-1] - ang_rel[0]]
        dd1 = bragg[0] + np.r_[0, FW]
        dd2 = bragg[0] + np.r_[0, ang_rel[1] - ang_rel[0]]

        dlamb_max = np.diff(coll.get_crystal_bragglamb(key=kspectro, bragg=dd0)[1])
        dlamb_mh = np.diff(coll.get_crystal_bragglamb(key=kspectro, bragg=dd1)[1])
        dlamb_res = np.diff(coll.get_crystal_bragglamb(key=kspectro, bragg=dd2)[1])

        if verb is True and res_lamb is not None:
            msg = (
                "Recommended res_lamb to ensure rocking curve overlap:\n"
                f"\t- edge-edge: \t{dlamb_max[0]:.2e}\n"
                f"\t- MH-to-MH: \t{dlamb_mh[0]:.2e}\n"
                f"\t- resolution: \t{dlamb_res[0]:.2e}\n"
                f"\tProvided: \t{dlamb:.2e}\n"
            )
            print(msg)

    return (
        nlamb,
        lamb,
        dlamb,
        pow_interp,
        ang_rel,
        dang,
        bragg,
    )


# ################################################
# ################################################
#           Loops
# ################################################


def _loop0(
    iok=None,
    ptsx=None,
    ptsy=None,
    ptsz=None,
    lpoly_post=None,
    min_threshold=None,
    # ref
    cent=None,
    nin=None,
    e0=None,
    e1=None,
    # p0
    p0x=None,
    p0y=None,
    p0z=None,
    # dang
    pix_size=None,
    dist_to_cam=None,
    dang=None,
    phi=None,
    shape_cam=None,
    # lamb
    lamb=None,
    dlamb=None,
    rocking_curve=None,
    bragg_per_pix=None,
    ang_rel=None,
    pow_interp=None,
    bragg=None,
    # optional
    n0=None,
    n1=None,
    # functions
    ptsvect_plane=None,
    coords_x01toxyz_plane=None,
    ptsvect_spectro=None,
    ptsvect_cam=None,
    # others
    ncounts=None,
    cbin0=None,
    cbin1=None,
    append=None,
    verb=True,
    **kwdargs,
):

    # ----------
    # prepare

    # counts, ph_count
    ncounts = np.full(shape_cam, 0.)
    sang = np.full(tuple(np.r_[shape_cam, ptsx.size]), 0.)
    sang0 = np.full(tuple(np.r_[shape_cam, ptsx.size]), 0.)
    sang_lamb = np.full(shape_cam, 0.)

    lsang = None
    if append is True:
        lp0, lp1 = [], []
        lpx, lpy, lpz = [], [], []
        lsang = []
        if lamb is None:
            dpow = None
        else:
            dpow = {ll: [] for ll in lamb}
    else:
        lp0, lp1 = None, None
        lpx, lpy, lpz = None, None, None
        dpow = None
        lsang = None

    # ----------
    # loop

    pti = np.full((3,), np.nan)
    npts0 = iok.sum()
    for ii, i0 in enumerate(iok.nonzero()[0]):

        pti[:] = np.r_[ptsx[i0], ptsy[i0], ptsz[i0]]

        # ------------------------------------------
        # initial polygon (crystal on its own plane)

        p0, p1 = ptsvect_plane(
            pts_x=ptsx[i0],
            pts_y=ptsy[i0],
            pts_z=ptsz[i0],
            vect_x=p0x - ptsx[i0],
            vect_y=p0y - ptsy[i0],
            vect_z=p0z - ptsz[i0],
            strict=True,
            return_x01=True,
        )[-2:]
        p_a = plg.Polygon(np.array([p0, p1]).T)

        if len(lpoly_post) > 0:

            # get equivalent aperture
            p0, p1 = _equivalent_apertures._get_equivalent_aperture(
                p_a=p_a,
                pt=pti,
                nop_pre=len(lpoly_post),
                lpoly_pre=lpoly_post,
                ptsvect=ptsvect_plane,
                min_threshold=min_threshold,
                # debug
                debug=False,
            )

            # skip if no intersection
            if p0 is None or p0.size == 0:
                continue

        # compute image
        (
            x0c, x1c, angles, dsang, cosi, iok,
            dangmin_str, x0if, x1if,
            ptsx1, ptsy1, ptsz1,
            ptsx2, ptsy2, ptsz2,
        ) = _get_points_on_camera_from_pts(
            p0=p0,
            p1=p1,
            pti=pti,
            # ref
            cent=cent,
            nin=nin,
            e0=e0,
            e1=e1,
            # dang
            pix_size=pix_size,
            dist_to_cam=dist_to_cam,
            dang=dang,
            phi=phi[ii],
            # optional
            n0=n0,
            n1=n1,
            # functions
            coords_x01toxyz_plane=coords_x01toxyz_plane,
            ptsvect_spectro=ptsvect_spectro,
            ptsvect_cam=ptsvect_cam,
            # debug
            debug=npts0 == 1,
        )

        if verb is True:
            msg = (
                f"\t\t{ii} / {npts0} pts \t dangmin: {dangmin_str}"
            )
            print(msg, end='\r')

        # stack coordinates
        if append is True:
            npts = iok.sum()
            lpx.append(np.array([
                np.full((npts,), ptsx[i0]),
                ptsx1[iok],
                ptsx2[iok],
            ]))
            lpy.append(np.array([
                np.full((npts,), ptsy[i0]),
                ptsy1[iok],
                ptsy2[iok],
            ]))
            lpz.append(np.array([
                np.full((npts,), ptsz[i0]),
                ptsz1[iok],
                ptsz2[iok],
            ]))

            # p0, p1
            lp0.append(x0c[iok])
            lp1.append(x1c[iok])

            # pow
            if lamb is not None:
                for kk, ll in enumerate(lamb):
                    dpow[ll].append(pow_interp(angles[iok] - bragg[kk]))

        # safety check
        iok2 = (
            (x0c[iok] >= cbin0[0])
            & (x0c[iok] <= cbin0[-1])
            & (x1c[iok] >= cbin1[0])
            & (x1c[iok] <= cbin1[-1])
        )

        if not np.any(iok2):
            continue

        # update index
        iok[iok] = iok2

        # 2d pixel by binning
        out = scpstats.binned_statistic_2d(
            x0c[iok],
            x1c[iok],
            None,
            statistic='count',
            bins=(cbin0, cbin1),
            expand_binnumbers=True,
        )

        ipixok = out.statistic > 0
        ncounts[ipixok] += out.statistic[ipixok]

        sang0[ipixok, i0] += scpstats.binned_statistic_2d(
            x0c[iok],
            x1c[iok],
            dsang[iok],
            statistic='sum',
            bins=(cbin0, cbin1),
            expand_binnumbers=True,
        ).statistic[ipixok]

        # taking into account rocking curve
        angles = angles[iok]
        dsang = dsang[iok]

        ip0, ip1 = ipixok.nonzero()
        indi = np.zeros((out.binnumber.shape[1],), dtype=bool)
        indj = np.zeros((out.binnumber.shape[1],), dtype=bool)
        for ii in np.unique(ip0):
            indi[:] = (out.binnumber[0, :] == (ii + 1))
            for jj in np.unique(ip1[ip0 == ii]):

                # indices
                indj[:] = indi & (out.binnumber[1, :] == jj + 1)

                if lamb is None:
                    # solid angle
                    sang[ii, jj, i0] += np.sum(dsang[indj])

                else:
                    # lambref
                    arelj = angles[indj] - bragg_per_pix[ii, jj]

                    # solid angle
                    sang[ii, jj, i0] += np.sum(
                        pow_interp(arelj)
                        * dsang[indj]
                    )

                    # ----------
                    # sang_lamb

                    angj = angles[indj]
                    ilamb = (
                        (angj[:, None] - bragg >= ang_rel[0])
                        & (angj[:, None] - bragg < ang_rel[-1])
                    )

                    if not np.any(ilamb):
                        continue

                    ilamb_n = np.any(ilamb, axis=0).nonzero()[0]

                    # binning of angles
                    for kk in ilamb_n:
                        sang_lamb[ii, jj] += np.sum(
                            pow_interp(angj[ilamb[:, kk]] - bragg[kk])
                            * dsang[indj][ilamb[:, kk]]
                        )

    # adjust sang_lamb
    if lamb is None:
        sang_lamb = np.nansum(sang0, axis=-1)

    return {
        'ncounts': {'data': ncounts, 'units': ''},
        'sang0': {'data': sang0, 'units': 'sr'},
        'sang': {'data': sang, 'units': 'sr'},    # not really useful?
        'sang_lamb': {'data': sang_lamb, 'units': 'sr'},
        'lamb': {'data': lamb, 'units': 'm'},
        'dlamb': {'data': dlamb, 'units': 'm'},
        'lp0': {'data': lp0, 'units': 'm'},
        'lp1': {'data': lp1, 'units': 'm'},
        'lpx': {'data': lpx, 'units': 'm'},
        'lpy': {'data': lpy, 'units': 'm'},
        'lpz': {'data': lpz, 'units': 'm'},
        'dpow': {'data': dpow, 'units': ''},
        'lsang': {'data': lsang, 'units': 'sr'},
    }


def _loop1(
    # specific
    coll=None,
    key=None,
    key_cam=None,
    key_mesh=None,
    margin_poly=None,
    res_phi=None,
    dsamp=None,
    # common
    iok=None,
    ptsx=None,
    ptsy=None,
    ptsz=None,
    lpoly_post=None,
    min_threshold=None,
    # ref
    cent=None,
    nin=None,
    e0=None,
    e1=None,
    # p0
    p0x=None,
    p0y=None,
    p0z=None,
    # dang
    pix_size=None,
    dist_to_cam=None,
    dang=None,
    phi=None,
    shape_cam=None,
    # optional
    n0=None,
    n1=None,
    # functions
    ptsvect_plane=None,
    coords_x01toxyz_plane=None,
    ptsvect_spectro=None,
    ptsvect_cam=None,
    # others
    ncounts=None,
    cbin0=None,
    cbin1=None,
    # others
    verb=True,
    **kwdargs,
):

    # --------------------------
    # get overall polygons

    pcross0, pcross1 = _utilities._get_overall_polygons(
        coll=coll,
        doptics=coll.dobj['diagnostic'][key]['doptics'],
        key_cam=key_cam,
        poly='pcross',
    )

    phor0, phor1 = _utilities._get_overall_polygons(
        coll=coll,
        doptics=coll.dobj['diagnostic'][key]['doptics'],
        key_cam=key_cam,
        poly='phor',
    )

    # --------------------------
    # add margins

    pcross0, pcross1 = _utilities._get_poly_margin(
        # polygon
        p0=pcross0,
        p1=pcross1,
        # margin
        margin=margin_poly,
    )

    phor0, phor1 = _utilities._get_poly_margin(
        # polygon
        p0=phor0,
        p1=phor1,
        # margin
        margin=margin_poly,
    )

    # ------------------------
    # get ind in cross-section

    sh = dsamp['x0']['data'].shape
    x0u = dsamp['x0']['data'][:, 0]
    x1u = dsamp['x1']['data'][0, :]
    x0f = dsamp['x0']['data'].ravel()
    x1f = dsamp['x1']['data'].ravel()

    pcross = Path(np.array([pcross0, pcross1]).T)
    ind = (
        dsamp['ind']['data']
        & pcross.contains_points(np.array([x0f, x1f]).T).reshape(sh)
    )
    nRZ = ind.sum()

    # R and Z indices
    ir, iz = ind.nonzero()
    iru = np.unique(ir)

    # ----------
    # get dphi_r

    phi_hor = np.arctan2(phor1, phor0)
    phimin = np.nanmin(phi_hor)
    phimax = np.nanmax(phi_hor)

    # get dphi vs phor
    dphi_r = _utilities._get_dphi_from_R_phor(
        R=x0u[iru],
        phor0=phor0,
        phor1=phor1,
        phimin=phimin,
        phimax=phimax,
        res=res_phi,
        out=True,
    )

    # --------------
    # prepare output

    indr = np.zeros((nRZ,), dtype=int)
    indz = np.zeros((nRZ,), dtype=int)
    ncounts = np.full(shape_cam, 0.)

    lp0, lp1 = [], []
    xx, yy = [], []

    dr = np.mean(np.diff(x0u))
    dz = np.mean(np.diff(x1u))
    ipts = 0
    pti = np.r_[0., 0., 0.]
    nru = iru.size
    for i00, i0 in enumerate(iru):

        indiz = ir == i0
        nz = indiz.sum()
        if np.all(np.isnan(dphi_r[:, i00])):
            ipts += nz
            continue

        nphi = np.ceil(x0u[i0]*(dphi_r[1, i00] - dphi_r[0, i00]) / res_phi).astype(int)
        phir = np.linspace(dphi_r[0, i00], dphi_r[1, i00], nphi)
        cosphi = np.cos(phir)
        sinphi = np.sin(phir)

        dphi = phir[1] - phir[0]
        dv = dr * x0u[i0] * dphi * dz

        for i11, i1 in enumerate(iz[indiz]):

            indr[ipts] = i0
            indz[ipts] = i1
            pti[2] = x1u[i1]
            done_xy = []

            for i2, phii in enumerate(phir):

                # set point
                pti[0] = x0u[i0]*cosphi[i2]
                pti[1] = x0u[i0]*sinphi[i2]

                # ------------------------------------------
                # initial polygon (crystal on its own plane)

                p0, p1 = ptsvect_plane(
                    pts_x=pti[0],
                    pts_y=pti[1],
                    pts_z=pti[2],
                    vect_x=p0x - pti[0],
                    vect_y=p0y - pti[1],
                    vect_z=p0z - pti[2],
                    strict=True,
                    return_x01=True,
                )[-2:]
                p_a = plg.Polygon(np.array([p0, p1]).T)

                if len(lpoly_post) > 0:
                    # get equivalent aperture
                    p0, p1 = _equivalent_apertures._get_equivalent_aperture(
                        p_a=p_a,
                        pt=pti,
                        nop_pre=len(lpoly_post),
                        lpoly_pre=lpoly_post,
                        ptsvect=ptsvect_plane,
                        min_threshold=min_threshold,
                    )

                    # skip if no intersection
                    if p0 is None or p0.size == 0:
                        continue

                # compute image
                (
                    x0c, x1c, angles, dsang, cosi, iok,
                    dangmin_str, x0if, x1if,
                ) = _get_points_on_camera_from_pts(
                    p0=p0,
                    p1=p1,
                    pti=pti,
                    # ref
                    cent=cent,
                    nin=nin,
                    e0=e0,
                    e1=e1,
                    # dang
                    pix_size=pix_size,
                    dist_to_cam=dist_to_cam,
                    dang=dang,
                    phi=phii,
                    # resoluions
                    n0=n0,
                    n1=n1,
                    # functions
                    coords_x01toxyz_plane=coords_x01toxyz_plane,
                    ptsvect_spectro=ptsvect_spectro,
                    ptsvect_cam=ptsvect_cam,
                )[:9]

                # update
                lp0.append(x0c)
                lp1.append(x1c)
                if i2 not in done_xy:
                    xx.append(x0u[i0] * np.cos(phii))
                    yy.append(x0u[i0] * np.sin(phii))
                    done_xy.append(i2)

                if verb is True:
                    msg = (
                        f"\t\t{i00} / {nru}, {i11} / {nz}, {i2} / {nphi}"
                        f":  {iok.sum()} pts   "
                        f"\t dangmin: {dangmin_str}"
                    )
                    print(msg, end='\r')

                # safety check
                iok2 = (
                    (x0c[iok] >= cbin0[0])
                    & (x0c[iok] <= cbin0[-1])
                    & (x1c[iok] >= cbin1[0])
                    & (x1c[iok] <= cbin1[-1])
                )

                if not np.any(iok2):
                    continue

                # update index
                iok[iok] = iok2

                # 2d pixel by binning
                out = scpstats.binned_statistic_2d(
                    x0c[iok],
                    x1c[iok],
                    dsang[iok],
                    statistic='sum',
                    bins=(cbin0, cbin1),
                    expand_binnumbers=True,
                )

                ipixok = out.statistic > 0
                ncounts[ipixok] += out.statistic[ipixok] * dv

    # ------------------
    # length for etendue

    # mesh envelop
    dout = coll.get_mesh_outline(key=key_mesh)
    p0 = dout['x0']['data']
    p1 = dout['x1']['data']

    etendue, length = _get_etendue_length(
        coll=coll,
        doptics=coll.dobj['diagnostic'][key]['doptics'][key_cam],
        poly=np.array([p0, p1]),
    )

    return {
        'ncounts': {'data': ncounts / length, 'units': 'm3/m'},
        'lp0': {'data': lp0, 'units': 'm'},
        'lp1': {'data': lp1, 'units': 'm'},
        'ptsx': {'data': np.array(xx), 'units': 'm'},
        'ptsy': {'data': np.array(yy), 'units': 'm'},
        'ptsr': {'data': x0u[ir], 'units': 'm'},
        'ptsz': {'data': x1u[iz], 'units': 'm'},
    }


def _get_etendue_length(
    coll=None,
    doptics=None,
    poly=None,
):

    # ----------------------
    # get etendue and length

    ketend = doptics['etendue']
    etendue = coll.ddata[ketend]['data']

    # -------------------------
    # los through mesh envelopp

    # los
    klos = doptics['los']
    ptsx, ptsy, ptsz = coll.get_rays_pts(key=klos)
    vectx, vecty, vectz = coll.get_rays_vect(key=klos)

    iok = np.isfinite(vectx[-1, ...])
    length = np.full(vectx.shape[1:], np.nan)

    DD = np.array([
        ptsx[-2, ...][iok],
        ptsy[-2, ...][iok],
        ptsz[-2, ...][iok],
    ])
    uu = np.array([
        vectx[-1, ...][iok],
        vecty[-1, ...][iok],
        vectz[-1, ...][iok],
    ])

    # Prepare structures
    from ..geom import _core
    ves = _core.Ves(
        Poly=poly,
        Name='temp',
        Exp='',
    )

    conf = _core.Config(
        lStruct=[ves],
        Name='temp',
        Exp='',
    )

    # ray-tracing
    cam = _core.CamLOS1D(
        dgeom=(DD, uu),
        config=conf,
        Name='temp',
        Diag='',
        Exp='',
        strict=False,
    )

    # length
    length[iok] = (cam.dgeom['kOut'] - cam.dgeom['kIn'])

    return etendue, length


# ################################################
# ################################################
#           Sub-routine
# ################################################


def _get_points_on_camera_from_pts(
    p0=None,
    p1=None,
    pti=None,
    # ref
    cent=None,
    nin=None,
    e0=None,
    e1=None,
    # dang
    pix_size=None,
    dist_to_cam=None,
    dang=None,
    phi=None,
    # optional
    n0=None,
    n1=None,
    # functions
    coords_x01toxyz_plane=None,
    ptsvect_spectro=None,
    ptsvect_cam=None,
    # output
    ptsx_all=None,
    ptsy_all=None,
    ptsz_all=None,
    # debug
    debug=None,
):

    # anuglar resolution associated to pixels
    vect = cent - pti
    dist = np.linalg.norm(vect)
    vect = vect / dist
    dang_pix = pix_size / (dist_to_cam + dist)

    # dang
    dang_min = min(dang, 0.25*dang_pix)
    dangmin_str = f"rock {dang:.2e} vs {0.25*dang_pix:.2e} 1/4 pixel"

    # set n0, n1
    p0min, p0max = p0.min(), p0.max()
    p1min, p1max = p1.min(), p1.max()

    if n0 is None:
        cos0 = np.linalg.norm(np.cross(e0, vect))
        ang0 = cos0 * (p0max - p0min) / dist
        n0 = int(np.ceil(ang0 / dang_min)) + 2
    if n1 is None:
        cos1 = np.linalg.norm(np.cross(e1, vect))
        ang1 = cos1 * (p1max - p1min) / dist
        n1 = int(np.ceil(ang1 / dang_min)) + 2

    # sample 2d equivalent aperture
    margin0 = 1e-6 * (p0max - p0min)
    margin1 = 1e-6 * (p1max - p1min)
    x0i = np.linspace(p0min + margin0, p0max - margin0, n0)
    x1i = np.linspace(p1min + margin1, p1max - margin1, n1)
    dx0 = x0i[1] - x0i[0]
    dx1 = x1i[1] - x1i[0]

    # mesh
    x0if = np.repeat(x0i[:, None], n1, axis=1)
    x1if = np.repeat(x1i[None, :], n0, axis=0)
    ind = Path(np.array([p0, p1]).T).contains_points(
        np.array([x0if.ravel(), x1if.ravel()]).T
    ).reshape((n0, n1))

    # back to 3d
    xx, yy, zz = coords_x01toxyz_plane(
        x0=x0if,
        x1=x1if,
    )

    # approx solid angles
    ax = xx - pti[0]
    ay = yy - pti[1]
    az = zz - pti[2]
    di = np.sqrt(ax**2 + ay**2 + az**2)
    cos = np.abs((nin[0] * ax + nin[1] * ay + nin[2] * az) / di)
    dsang = dx0 * dx1 * cos / di**2

    # get normalized vector from plasma point to crystal
    vectx = ax / di
    vecty = ay / di
    vectz = az / di

    # get local cosine vs toroidal direction (for doppler)
    cos = -vectx*np.sin(phi) + vecty*np.cos(phi)

    # get reflexion
    (
        ptsx1, ptsy1, ptsz1,
        vx, vy, vz,
        angles,
    ) = ptsvect_spectro(
        pts_x=pti[0],
        pts_y=pti[1],
        pts_z=pti[2],
        vect_x=vectx,
        vect_y=vecty,
        vect_z=vectz,
        strict=False,
        return_x01=False,
    )[:7]

    # get x0, x1 on camera
    x0c, x1c, ptsx2, ptsy2, ptsz2 = ptsvect_cam(
        pts_x=ptsx1,
        pts_y=ptsy1,
        pts_z=ptsz1,
        vect_x=vx,
        vect_y=vy,
        vect_z=vz,
    )

    return (
        x0c, x1c, angles, dsang, cos, ind,
        dangmin_str, x0if, x1if,
        ptsx1, ptsy1, ptsz1,
        ptsx2, ptsy2, ptsz2,
    )


# ####################################################
# ####################################################
#               plot
# ####################################################


def _plot(
    coll=None,
    key=None,
    key_cam=None,
    dout=None,
    # auxiliary
    cbin0=None,
    cbin1=None,
    is2d=None,
    # options
    plot_config=None,
    plot_pixels=None,
    vmin=None,
    vmax=None,
    aspect3d=None,
    elements=None,
    colorbar=None,
    # options
    dax=None,
    fs=None,
    dmargin=None,
    wintit=None,
):

    # --------------
    # check inputs

    aspect3d = ds._generic_check._check_var(
        aspect3d, 'aspect3d',
        default='equal',
        allowed=['auto', 'equal'],
    )

    elements = ds._generic_check._check_var(
        elements, 'elements',
        default='o',
        types=str,
    )

    # --------------
    # add ncounts to ddata

    ktemp = f'{key_cam}_ncounts'
    nn = len([
        k0 for k0 in coll.ddata.keys()
        if k0.startswith(ktemp)
    ])
    if nn > 0:
        ktemp = f'{ktemp}{nn}'

    coll.add_data(
        key=ktemp,
        ref=coll.dobj['camera'][key_cam]['dgeom']['ref'],
        **dout['ncounts'],
    )

    kd = 'sang_lamb'
    if vmin is None:
        vmin = 0
    if vmax is None:
        vmax = np.nanmax(dout[kd]['data'])

    # --------------
    # prepare

    dplot = coll.get_diagnostic_dplot(
        key=key,
        key_cam=key_cam,
        optics=None,
        elements=elements,
    )

    # --------------
    # plot_pixels

    out0, out1 = coll.get_optics_outline(key_cam, total=True)
    ck0f = np.array([cbin0, cbin0, np.full((cbin0.size,), np.nan)])
    ck1f = np.array([cbin1, cbin1, np.full((cbin1.size,), np.nan)])
    ck01 = np.r_[np.min(cbin1), np.max(cbin1), np.nan]
    ck10 = np.r_[np.min(cbin0), np.max(cbin0), np.nan]

    # --------------
    # pepare dax

    if dax is None:
        dax, cax = _get_dax(
            fs=fs,
            dmargin=dmargin,
        )
        # dax = _generic_plot.get_dax_diag(
            # proj=['cross', 'hor', '3d', 'camera'],
            # dmargin=dmargin,
            # fs=fs,
            # wintit=wintit,
            # tit=key,
            # is2d=True,
            # key_cam=['rays', 'binned'],
        # )

    dax = _generic_check._check_dax(dax=dax, main='cross')

    # --------------
    # prepare camera

    kax = 'cam_all'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        ax.plot(
            np.r_[out0, out0[0]],
            np.r_[out1, out1[0]],
            c='k',
            ls='-',
            lw=0.5,
        )

        if plot_pixels is True:
            ax.plot(
                ck0f.T.ravel(),
                np.tile(ck01, cbin0.size),
                c='k',
                ls='-',
                lw=0.1,
            )
            ax.plot(
                np.tile(ck10, cbin1.size),
                ck1f.T.ravel(),
                c='k',
                ls='-',
                lw=0.1,
            )


    # -------------
    # mesh sampling

    if dout.get('lpx') is None:

        kax = 'cross'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']

            ax.plot(
                dout['ptsr']['data'],
                dout['ptsz']['data'],
                ls='None',
                lw=1.,
                marker='.',
            )

        kax = 'hor'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']

            ax.plot(
                dout['ptsx']['data'],
                dout['ptsy']['data'],
                ls='None',
                lw=1.,
                marker='.',
            )

        kax = 'cam_all'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']

            p0 = np.concatenate([pp.ravel() for pp in dout['lp0']['data']])
            p1 = np.concatenate([pp.ravel() for pp in dout['lp1']['data']])
            ax.plot(
                p0,
                p1,
                ls='None',
                marker='.',
            )

    else:

        # --------------
        # ray-tracing

        for ii in range(len(dout['lp0']['data'])):

            p0 = dout['lp0']['data'][ii]
            p1 = dout['lp1']['data'][ii]

            # --------------
            # camera image

            kax = 'cam_all'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                if dout['lamb']['data'] is None:
                    ax.plot(
                        p0,
                        p1,
                        ls='None',
                        marker='.',
                        # c=dout['lcolor'][ii],
                    )
                elif dout['rocking_curve'] is False:
                    ax.plot(
                        p0[dout['lindin'][ii]],
                        p1[dout['lindin'][ii]],
                        ls='None',
                        marker='.',
                        # c=dout['lcolor'][ii],
                    )
                    ax.plot(
                        p0[~dout['lindin'][ii]],
                        p1[~dout['lindin'][ii]],
                        ls='None',
                        marker='.',
                        c=(0.8, 0.8, 0.8),
                    )
                else:
                    ax.scatter(
                        p0,
                        p1,
                        # c=dout['lcolor'][ii],
                        s=8,
                        marker='.',
                    )

            # --------------
            # rays

            ind = np.arange(3, dout['lpx']['data'][ii].size, 3)
            px = np.insert(dout['lpx']['data'][ii].T.ravel(), ind, np.nan)
            py = np.insert(dout['lpy']['data'][ii].T.ravel(), ind, np.nan)
            pz = np.insert(dout['lpz']['data'][ii].T.ravel(), ind, np.nan)

            kax = 'cross'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                ax.plot(
                    np.hypot(px, py),
                    pz,
                    ls='-',
                    lw=1.,
                )

            kax = 'hor'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                ax.plot(
                    px,
                    py,
                    ls='-',
                    lw=1.,
                )

            kax = '3d'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                ax.plot(
                    px,
                    py,
                    pz,
                    ls='-',
                    lw=1.,
                )

    # ----------
    # counts

    kax = 'cam_bin'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        extent = (cbin0[0], cbin0[-1], cbin1[0], cbin1[-1])
        im = ax.imshow(
            dout[kd]['data'].T,
            extent=extent,
            interpolation='nearest',
            origin='lower',
            vmin=vmin,
            vmax=vmax,
        )

        ax.set_title(kd, size=12, fontweight='bold')
        if colorbar is True:
            cb = plt.colorbar(im, ax=ax, cax=cax)
            cb.set_label(dout[kd]['units'], size=12, weight='bold')

    # ----------
    # diag geom

    _class8_plot._plot_diag_geom(
        dax=dax,
        key_cam=key_cam,
        dplot=dplot,
        is2d=coll.dobj['diagnostic'][key]['is2d'],
    )

    # -------
    # config

    if plot_config.__class__.__name__ == 'Config':

        kax = 'cross'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj=kax, dLeg=False)

        kax = 'hor'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj=kax, dLeg=False)

    # --------------
    # aspect3d

    if aspect3d == 'equal' and dax.get('3d') is not None:
        ds.set_aspect3d(ax=dax['3d']['handle'])

    # --------------
    # clean up

    coll.remove_data(ktemp)

    return


def _get_dax(
    fs=None,
    dmargin=None,
):

    # ----------
    # check

    if fs is None:
        fs = (15, 9)

    dmargin = ds._generic_check._check_var(
        dmargin, 'dmargin',
        types=dict,
        default={
            'bottom': 0.05, 'top': 0.95,
            'left': 0.05, 'right': 0.94,
            'wspace': 0.20, 'hspace': 0.25,
        },
    )

    # --------
    # prepare

    ncols = 31
    nh = int((ncols - 1) / 2)
    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(ncols=ncols, nrows=4, **dmargin)

    # --------
    # create

    ax0 = fig.add_subplot(gs[:2, :nh], aspect='equal', adjustable='datalim')
    ax0.set_xlabel('X (m)')
    ax0.set_ylabel('Y (m)')

    ax1 = fig.add_subplot(gs[2:, :nh], aspect='equal', adjustable='datalim')
    ax1.set_xlabel('R (m)')
    ax1.set_ylabel('Z (m)')

    ax2 = fig.add_subplot(gs[:2, nh:-1], projection='3d')
    ax2.set_xlabel('X (m)')
    ax2.set_ylabel('Y (m)')
    ax2.set_ylabel('Z (m)')

    ax3 = fig.add_subplot(gs[2, nh:-1], aspect='equal')
    ax3.set_ylabel('x1 (m)')

    ax4 = fig.add_subplot(gs[3, nh:-1], sharex=ax3, sharey=ax3)
    ax4.set_ylabel('x1 (m)')
    ax4.set_xlabel('x0 (m)')

    cax = fig.add_subplot(gs[3, -1])

    dax = {
        'cross': {'handle': ax0, 'type': 'cross'},
        'hor': {'handle': ax1, 'type': 'hor'},
        '3d': {'handle': ax2, 'type': '3d'},
        'cam_all': {'handle': ax3, 'type': 'camera'},
        'cam_bin': {'handle': ax4, 'type': 'camera'},
    }

    return dax, cax


# -*- coding: utf-8 -*-
"""
Provide data handling class and methods (storing, processing, plotting...)
"""
from ._core import *
from ._DataCollection_class1_interactivity import *
from ._class01_eqdsk import *
from ._spectrallines_class import *
from ._class10_algos import get_available_inversions_algo
from ._class10_Inversion import Inversion as Collection
from ._spectralunits import *
from ._saveload import *
from ._class08_saveload_from_file import load as load_diagnostic_from_file


# -*- coding: utf-8 -*-


import datetime as dtm      # DB
import numpy as np
import scipy.stats as scpstats
from matplotlib.path import Path
import matplotlib.pyplot as plt       # DB
import matplotlib.gridspec as gridspec
import Polygon as plg


from . import _class8_equivalent_apertures as _equivalent_apertures
from . import _class8_vos_utilities as _utilities
from . import _class8_reverse_ray_tracing as _reverse_rt


# ###########################################################
# ###########################################################
#               Main
# ###########################################################


def _vos(
    # ressources
    coll=None,
    doptics=None,
    key_diag=None,
    key_cam=None,
    dsamp=None,
    # inputs
    x0u=None,
    x1u=None,
    x0f=None,
    x1f=None,
    x0l=None,
    x1l=None,
    # overall polygons
    pcross0=None,
    pcross1=None,
    phor0=None,
    phor1=None,
    dphi_r=None,
    sh=None,
    res_phi=None,
    lamb=None,
    res_lamb=None,
    res_rock_curve=None,
    n0=None,
    n1=None,
    convexHull=None,
    bool_cross=None,
    # parameters
    min_threshold=None,
    margin_poly=None,
    visibility=None,
    verb=None,
    # debug
    debug=None,
    # timing
    timing=None,
    dt11=None,
    dt111=None,
    dt1111=None,
    dt2222=None,
    dt3333=None,
    dt4444=None,
    dt222=None,
    dt333=None,
    dt22=None,
    # unused
    **kwdargs,
):
    """ vos computation for spectrometers """

    if timing:
        t00 = dtm.datetime.now()     # DB

    # -----------------
    # prepare optics

    (
        kspectro,
        lpoly_post,
        p0x, p0y, p0z,
        nin, e0, e1,
        ptsvect_plane,
        ptsvect_spectro,
        ptsvect_cam,
        coords_x01toxyz_plane,
        cent_spectro,
        cent_cam,
        dist_to_cam,
        pix_size,
        cbin0, cbin1,
    ) = _reverse_rt._prepare_optics(
        coll=coll,
        key=key_diag,
        key_cam=key_cam,
    )

    # --------------------------
    # get overall polygons

    pcross0, pcross1 = _utilities._get_overall_polygons(
        coll=coll,
        doptics=coll.dobj['diagnostic'][key_diag]['doptics'],
        key_cam=key_cam,
        poly='pcross',
        convexHull=convexHull,
    )

    phor0, phor1 = _utilities._get_overall_polygons(
        coll=coll,
        doptics=coll.dobj['diagnostic'][key_diag]['doptics'],
        key_cam=key_cam,
        poly='phor',
        convexHull=convexHull,
    )

    # --------------------------
    # add margins

    pcross0, pcross1 = _utilities._get_poly_margin(
        # polygon
        p0=pcross0,
        p1=pcross1,
        # margin
        margin=margin_poly,
    )

    phor0, phor1 = _utilities._get_poly_margin(
        # polygon
        p0=phor0,
        p1=phor1,
        # margin
        margin=margin_poly,
    )

    # ------------------------
    # get ind in cross-section

    pcross = Path(np.array([pcross0, pcross1]).T)
    ind = (
        dsamp['ind']['data']
        & pcross.contains_points(np.array([x0f, x1f]).T).reshape(sh)
    )
    nRZ = ind.sum()

    # R and Z indices
    ir, iz = ind.nonzero()
    iru = np.unique(ir)

    # ----------
    # get dphi_r

    # dphi = doptics[key_cam]['dvos']['dphi']
    # phimin = np.nanmin(dphi[0, :])
    # phimax = np.nanmax(dphi[1, :])

    phi_hor = np.arctan2(phor1, phor0)
    phimin = np.nanmin(phi_hor)
    phimax = np.nanmax(phi_hor)

    # get dphi vs phor
    dphi_r = _utilities._get_dphi_from_R_phor(
        R=x0u[iru],
        phor0=phor0,
        phor1=phor1,
        phimin=phimin,
        phimax=phimax,
        res=res_phi,
        out=True,
    )

    # -------------------------------------
    # prepare lambda, angles, rocking_curve

    (
        nlamb,
        lamb,
        dlamb,
        pow_interp,
        ang_rel,
        dang,
        bragg,
    ) = _reverse_rt._prepare_lamb(
        coll=coll,
        key_diag=key_diag,
        key_cam=key_cam,
        kspectro=kspectro,
        lamb=lamb,
        res_lamb=res_lamb,
        res_rock_curve=res_rock_curve,
        verb=verb,
    )

    # --------------
    # prepare output

    shape_cam = coll.dobj['camera'][key_cam]['dgeom']['shape']
    is2d = len(shape_cam) == 2

    shape0 = tuple(np.r_[shape_cam, nRZ])
    ncounts = np.full(shape0, 0.)
    cos = np.full(shape0, 0.)
    phi_mean = np.full(shape0, 0.)
    phi_min = np.full(shape0, np.inf)
    phi_max = np.full(shape0, -np.inf)
    indr = np.zeros((nRZ,), dtype=int)
    indz = np.zeros((nRZ,), dtype=int)
    dV = np.full((nRZ,), np.nan)

    shape1 = tuple(np.r_[shape_cam, nRZ, nlamb])
    ph_count = np.full(shape1, 0.)

    etendlen = np.full(shape_cam, 0.)
    # ph_approx = np.full(shape1, 0.)
    # sang = np.full(shape1, 0.)
    # dang_rel = np.full(shape1, 0.)
    # nphi_all = np.full(shape1, 0.)
    # FW = coll.dobj[cls_spectro][kspectro]['dmat']['drock']['FW']
    # kp = coll.dobj[cls_spectro][kspectro]['dmat']['drock']['power_ratio']
    # POW = coll.ddata[kp]['data'].max()

    if timing:
        t11 = dtm.datetime.now()     # DB
        dt11 += (t11-t00).total_seconds()

    # ----------
    # verb

    if verb is True:
        msg = (
            f"\tlamb.shape: {lamb.shape}\n"
            f"\tang_rel.shape: {ang_rel.shape}\n"
            f"\tiru.size: {iru.size}\n"
            f"\tnRZ: {nRZ}\n"
        )
        print(msg)

    # ---------------------
    # loop in plasma points

    if debug is True:
        dx0 = {
            i0: {
                i1: [] for i1 in np.unique(iz[ir == i0])
            }
            for i0 in iru
        }
        dx1 = {
            i0: {
                i1: [] for i1 in np.unique(iz[ir == i0])
            }
            for i0 in iru
        }

    dr = np.mean(np.diff(x0u))
    dz = np.mean(np.diff(x1u))
    ipts = 0
    pti = np.r_[0., 0., 0.]
    nru = iru.size
    for i00, i0 in enumerate(iru):

        indiz = ir == i0
        nz = indiz.sum()
        if np.all(np.isnan(dphi_r[:, i00])):
            ipts += nz
            continue

        nphi = max(
            np.ceil(
                x0u[i0]*(dphi_r[1, i00] - dphi_r[0, i00]) / res_phi
            ).astype(int),
            3,
        )

        phir = np.linspace(dphi_r[0, i00], dphi_r[1, i00], nphi)
        cosphi = np.cos(phir)
        sinphi = np.sin(phir)

        dphi = phir[1] - phir[0]
        dv = dr * x0u[i0] * dphi * dz

        for i11, i1 in enumerate(iz[indiz]):

            indr[ipts] = i0
            indz[ipts] = i1
            dV[ipts] = dv
            pti[2] = x1u[i1]

            for i2, phii in enumerate(phir):

                if timing:
                    t000 = dtm.datetime.now()     # DB

                # set point
                pti[0] = x0u[i0] * cosphi[i2]
                pti[1] = x0u[i0] * sinphi[i2]
                # phi[ipts] = phii

                # ------------------------------------------
                # initial polygon (crystal on its own plane)

                p0, p1 = ptsvect_plane(
                    pts_x=pti[0],
                    pts_y=pti[1],
                    pts_z=pti[2],
                    vect_x=p0x - pti[0],
                    vect_y=p0y - pti[1],
                    vect_z=p0z - pti[2],
                    strict=True,
                    return_x01=True,
                )[-2:]
                p_a = plg.Polygon(np.array([p0, p1]).T)

                if len(lpoly_post) > 0:
                    # get equivalent aperture
                    p0, p1 = _equivalent_apertures._get_equivalent_aperture(
                        p_a=p_a,
                        pt=pti,
                        nop_pre=len(lpoly_post),
                        lpoly_pre=lpoly_post,
                        ptsvect=ptsvect_plane,
                        min_threshold=min_threshold,
                    )

                    # skip if no intersection
                    if p0 is None or p0.size == 0:
                        continue

                if timing:
                    t111 = dtm.datetime.now()     # DB
                    dt111 += (t111-t000).total_seconds()

                # compute image
                (
                    x0c, x1c, angles, dsang, cosi, iok,
                    dangmin_str, x0if, x1if,
                ) = _reverse_rt._get_points_on_camera_from_pts(
                    p0=p0,
                    p1=p1,
                    pti=pti,
                    # ref
                    cent=cent_spectro,
                    nin=nin,
                    e0=e0,
                    e1=e1,
                    # dang
                    pix_size=pix_size,
                    dist_to_cam=dist_to_cam,
                    dang=dang,
                    phi=phii,
                    # resoluions
                    n0=n0,
                    n1=n1,
                    # functions
                    coords_x01toxyz_plane=coords_x01toxyz_plane,
                    ptsvect_spectro=ptsvect_spectro,
                    ptsvect_cam=ptsvect_cam,
                )[:9]

                if verb is True:
                    msg = (
                        f"\t\t{i00} / {nru}, {i11} / {nz}, {i2} / {nphi}"
                        f":  {iok.sum()} pts   "
                        f"\t dangmin: {dangmin_str}"
                    )
                    print(msg, end='\r')

                if timing:
                    # dt1111, dt2222, dt3333, dt4444 = out
                    t222 = dtm.datetime.now()     # DB
                    dt222 += (t222-t111).total_seconds()

                # safety check
                iok2 = (
                    (x0c[iok] >= cbin0[0])
                    & (x0c[iok] <= cbin0[-1])
                    & (x1c[iok] >= cbin1[0])
                    & (x1c[iok] <= cbin1[-1])
                )

                # ---------- DEBUG ------------
                if debug is True:
                    # _plot_debug(
                        # coll=coll,
                        # key_cam=key_cam,
                        # cbin0=cbin0,
                        # cbin1=cbin1,
                        # x0c=x0c,
                        # x1c=x1c,
                        # cos=cosi,
                        # angles=angles,
                        # iok=iok,
                        # p0=p0,
                        # p1=p1,
                        # x0if=x0if,
                        # x1if=x1if,
                    # )
                    dx0[i0][i1].append(x0c)
                    dx1[i0][i1].append(x1c)
                # -------- END DEBUG ----------


                if not np.any(iok2):
                    continue

                # update index
                iok[iok] = iok2

                # 2d pixel by binning
                out = scpstats.binned_statistic_2d(
                    x0c[iok],
                    x1c[iok],
                    None,
                    statistic='count',
                    bins=(cbin0, cbin1),
                    expand_binnumbers=True,
                )

                ipixok = out.statistic > 0
                ncounts[ipixok, ipts] += out.statistic[ipixok]

                # temporary
                outsa = scpstats.binned_statistic_2d(
                    x0c[iok],
                    x1c[iok],
                    dsang[iok],
                    statistic='sum',
                    bins=(cbin0, cbin1),
                    expand_binnumbers=True,
                )
                etendlen[ipixok] += outsa.statistic[ipixok] * dv

                # adjust phimean
                phi_mean[ipixok, ipts] += phii * out.statistic[ipixok]

                cosi = cosi[iok]
                angles = angles[iok]
                dsang = dsang[iok]

                ip0, ip1 = ipixok.nonzero()
                indi = np.zeros((out.binnumber.shape[1],), dtype=bool)
                indj = np.zeros((out.binnumber.shape[1],), dtype=bool)
                for ii in np.unique(ip0):
                    indi[:] = (out.binnumber[0, :] == (ii + 1))
                    for jj in np.unique(ip1[ip0 == ii]):

                        # indices
                        indj[:] = indi & (out.binnumber[1, :] == jj + 1)

                        # phi_min, phi_max
                        phi_min[ii, jj, ipts] = min(phi_min[ii, jj, ipts], phii)
                        phi_max[ii, jj, ipts] = max(phi_max[ii, jj, ipts], phii)

                        # cos
                        cos[ii, jj, ipts] += np.sum(cosi[indj])

                        # ilamb
                        angj = angles[indj]
                        ilamb = (
                            (angj[:, None] - bragg >= ang_rel[0])
                            & (angj[:, None] - bragg < ang_rel[-1])
                        )

                        if not np.any(ilamb):
                            continue

                        ilamb_n = np.any(ilamb, axis=0).nonzero()[0]

                        # nphi_all  # DB
                        # nphi_all[ii, jj, ipts, ilamb_n] += 1

                        # if False:
                        # binning of angles
                        for kk in ilamb_n:
                            ph_count[ii, jj, ipts, kk] += np.sum(
                                pow_interp(angj[ilamb[:, kk]] - bragg[kk])
                                * dsang[indj][ilamb[:, kk]]
                            ) * dv

                            # inds = np.searchsorted(
                            #     angbragg[:, kk],
                            #     angj[ilamb[:, kk]],
                            # )

                            # # update power_ratio * solid angle
                            # ph_count[ii, jj, ipts, kk] += np.sum(
                            #     pow_ratio[inds]
                            #     * dsang[indj][ilamb[:, kk]]
                            # ) * dv

                            # ------  DEBUG -------
                            # ph_approx[ii, jj, ipts, kk] += (
                                # POW
                                # * FW
                                # * np.sum(dang1[0, np.any(iok, axis=0)])
                            # ) * dv

                            # # sang
                            # sang[ii, jj, ipts, kk] += np.sum(
                                # dsang[indj][ilamb[:, kk]]
                            # )

                            # # dang_rel
                            # dang_rel[ii, jj, ipts, kk] += np.sum(
                                # dsang[indj][ilamb[:, kk]]
                            # )

                if timing:
                    t333 = dtm.datetime.now()     # DB
                    dt333 += (t333-t222).total_seconds()

            # update index
            ipts += 1

    # multiply by dlamb
    # Now done during synthetic signal compute (binning vs interp)
    # ph_count *= dlamb

    if timing:
        t22 = dtm.datetime.now()     # DB

    # remove useless points
    iin = np.any(np.any(ncounts > 0, axis=0), axis=0)
    if not np.all(iin):
        ncounts = ncounts[:, :, iin]
        cos = cos[:, :, iin]
        phi_mean = phi_mean[:, :, iin]
        phi_min = phi_min[:, :, iin]
        phi_max = phi_max[:, :, iin]
        ph_count = ph_count[:, :, iin, :]
        indr = indr[iin]
        indz = indz[iin]
        dV = dV[iin]
        # DEBUG
        # sang = sang[:, :, iin]
        # ph_approx = ph_approx[:, :, iin, :]
        # dang_rel = dang_rel[:, :, iin, :]
        # nphi_all = nphi_all[:, :, iin, :]

    # remove useless lamb
    iin = ph_count > 0.
    ilamb = np.any(np.any(np.any(iin, axis=0), axis=0), axis=0)
    if not np.all(ilamb):
        ph_count = ph_count[..., ilamb]
        lamb = lamb[ilamb]
        # DEBUG
        # sang = sang[..., ilamb]
        # ph_approx = ph_approx[..., ilamb]
        # dang_rel = dang_rel[..., ilamb]
        # nphi_all = nphi_all[..., ilamb]

    # average cos and phi_mean
    iout = ncounts == 0
    cos[~iout] = cos[~iout] / ncounts[~iout]
    phi_mean[~iout] = phi_mean[~iout] / ncounts[~iout]

    # clear
    cos[iout] = np.nan
    phi_mean[iout] = np.nan
    phi_min[iout] = np.nan
    phi_max[iout] = np.nan
    # ph_count[iout, :] = np.nan
    # DEBUG
    # sang[iout, :] = np.nan
    # ph_approx[iout, :] = np.nan
    # dang_rel[iout, :] = np.nan
    # nphi_all[iout, :] = np.nan

    lresh = [1, 1, lamb.size]
    if is2d:
        lresh.insert(0, 1)
    # phtot = np.sum(ph_count, axis=(-1, -2))
    # lamb0 = np.sum(ph_count * lamb.reshape(lresh), axis=(-1, -2)) / phtot
    # dlamb0 = dlamb * phtot / np.max(np.sum(ph_count, axis=-2), axis=-1)

    # ------ DEBUG --------
    if debug is True:
        _plot_debug(
            coll=coll,
            key_cam=key_cam,
            cbin0=cbin0,
            cbin1=cbin1,
            dx0=dx0,
            dx1=dx1,
        )
    # ---------------------

    # ----------------
    # prepare output

    # ref
    knpts = f'{key_diag}_{key_cam}_vos_npts'
    knlamb = f'{key_diag}_{key_cam}_vos_nlamb'

    # data
    klamb = f'{key_diag}_{key_cam}_vos_lamb'
    kir = f'{key_diag}_{key_cam}_vos_ir'
    kiz = f'{key_diag}_{key_cam}_vos_iz'
    kph = f'{key_diag}_{key_cam}_vos_ph'
    kcos = f'{key_diag}_{key_cam}_vos_cos'
    knc = f"{key_diag}_{key_cam}_vos_nc"
    kphimin = f'{key_diag}_{key_cam}_vos_phimin'
    kphimax = f'{key_diag}_{key_cam}_vos_phimax'

    # optional data
    kdV = f'{key_diag}_{key_cam}_vos_dV'
    ketl = f"{key_diag}_{key_cam}_vos_etendl"
    # klamb0 = f'{key_cam}_vos_lamb0'
    # kdlamb = f'{key_cam}_vos_dlamb'
    kphimean = f'{key_diag}_{key_cam}_vos_phimean'

    refcam = coll.dobj['camera'][key_cam]['dgeom']['ref']
    ref = tuple(list(refcam) + [knpts])
    refph = tuple(list(ref) + [knlamb])

    # -------------------
    # format output

    # dref
    dref = {
        'npts': {
            'key': knpts,
            'size': indr.shape[-1],
        },
        'nlamb': {
            'key': knlamb,
            'size': lamb.size,
        },
    }


    # dout
    dout = {
        'pcross0': None,
        'pcross1': None,
        'phor0': None,
        'phor1': None,
        # lamb
        'lamb': {
            'key': klamb,
            'data': lamb,
            'ref': (knlamb,),
            'units': 'm',
            'dim': 'distance',
        },

        # coordinates
        'indr_cross': {
            'key': kir,
            'data': indr,
            'ref': (knpts,),
            'units': None,
            'dim': 'index',
        },
        'indz_cross': {
            'key': kiz,
            'data': indz,
            'ref': (knpts,),
            'units': None,
            'dim': 'index',
        },

        # data
        'cos': {
            'key': kcos,
            'data': cos,
            'ref': ref,
            'units': None,
            'dim': 'cos',
        },
        'ncounts': {
            'key': knc,
            'data': ncounts,
            'ref': ref,
            'units': None,
            'dim': 'counts',
        },
        'ph': {
            'key': kph,
            'data': ph_count,
            'ref': refph,
            'units': 'sr.m3',
            'dim': 'transfert',
        },
        'phi_min': {
            'key': kphimin,
            'data': phi_min,
            'ref': ref,
            'units': 'rad',
            'dim': 'angle',
        },
        'phi_max': {
            'key': kphimax,
            'data': phi_max,
            'ref': ref,
            'units': 'rad',
            'dim': 'angle',
        },

        # optional
        'phi_mean': {
            'key': kphimean,
            'data': phi_mean,
            'ref': ref,
            'units': 'rad',
            'dim': 'angle',
        },
        'dV': {
            'key': kdV,
            'data': dV,
            'ref': (knpts,),
            'units': 'm3',
            'dim': 'volume',
        },

        # debug
        'etendlen': {
            'key': ketl,
            'data': etendlen,
            'ref': refcam,
            'units': 'sr.m3',
            'dim': 'etend*len',
        },
    }

    if timing:
        t33 = dtm.datetime.now()
        dt22 += (t33 - t22).total_seconds()

    return (
        dout, dref,
        dt11, dt22,
        dt111, dt222, dt333,
        dt1111, dt2222, dt3333, dt4444,
    )


# ################################################
# ################################################
#           Debug plot
# ################################################


def _plot_debug(
    coll=None,
    key_cam=None,
    cbin0=None,
    cbin1=None,
    dx0=None,
    dx1=None,
    x0c=None,
    x1c=None,
    cos=None,
    angles=None,
    iok=None,
    p0=None,
    p1=None,
    x0if=None,
    x1if=None,
):

    out0, out1 = coll.get_optics_outline(key_cam, total=True)
    ck0f = np.array([cbin0, cbin0, np.full((cbin0.size,), np.nan)])
    ck1f = np.array([cbin1, cbin1, np.full((cbin1.size,), np.nan)])
    ck01 = np.r_[np.min(cbin1), np.max(cbin1), np.nan]
    ck10 = np.r_[np.min(cbin0), np.max(cbin0), np.nan]

    fig = plt.figure(figsize=(14, 8))
    if dx0 is None:
        ldata = [
            ('cos vs toroidal', cos),
            ('angles vs crystal', angles),
            ('iok', iok)
        ]

        for ii, v0 in enumerate(ldata):
            ax = fig.add_subplot(2, 2, ii + 1, aspect='equal')
            ax.set_title(v0[0], size=12, fontweight='bold')
            ax.set_xlabel('x0 (m)', size=12, fontweight='bold')
            ax.set_xlabel('x1 (m)', size=12, fontweight='bold')

            # grid
            ax.plot(np.r_[out0, out0[0]], np.r_[out1, out1[0]], '.-k')
            ax.plot(
                ck0f.T.ravel(),
                np.tile(ck01, cbin0.size),
                '-k',
            )
            ax.plot(
                np.tile(ck10, cbin1.size),
                ck1f.T.ravel(),
                '-k',
            )

            # points
            im = ax.scatter(
                x0c,
                x1c,
                c=v0[1],
                s=4,
                edgecolors='None',
                marker='o',
            )
            plt.colorbar(im, ax=ax)

        # projected polygon
        ax = fig.add_subplot(2, 2, 4, aspect='equal', adjustable='datalim')
        ax.set_title('Projected polygon', size=12, fontweight='bold')
        ax.set_xlabel('x0 (m)', size=12, fontweight='bold')
        ax.set_xlabel('x1 (m)', size=12, fontweight='bold')

        ax.plot(
            np.r_[p0, p0[0]],
            np.r_[p1, p1[0]],
            c='k',
            ls='-',
            lw=1.,
        )
        ax.plot(
            x0if[iok],
            x1if[iok],
            '.g',
        )
        ax.plot(
            x0if[~iok],
            x1if[~iok],
            '.r',
        )

    else:

        dmargin = {
            'bottom': 0.08, 'top': 0.95,
            'left': 0.10, 'right': 0.95,
            'wspace': 0.50, 'hspace': 0.10,
        }

        gs = gridspec.GridSpec(ncols=20, nrows=2, **dmargin)
        ax0 = fig.add_subplot(gs[0, :-1], aspect='equal')
        ax1 = fig.add_subplot(
            gs[1, :-1],
            sharex=ax0,
            sharey=ax0,
        )
        ax2 = fig.add_subplot(gs[1, -1])

        ax0.set_ylabel('x1 (m)', size=12, fontweight='bold')

        ax1.set_xlabel('x0 (m)', size=12, fontweight='bold')
        ax1.set_ylabel('x1 (m)', size=12, fontweight='bold')

        # grid
        ax0.plot(np.r_[out0, out0[0]], np.r_[out1, out1[0]], '.-k')
        ax0.plot(
            ck0f.T.ravel(),
            np.tile(ck01, cbin0.size),
            '-k',
        )
        ax0.plot(
            np.tile(ck10, cbin1.size),
            ck1f.T.ravel(),
            '-k',
        )

        # pre-concatenate
        for i0, v0 in dx0.items():
            for i1, v1 in v0.items():
                if len(v1) > 0:
                    dx0[i0][i1] = np.concatenate([vv.ravel() for vv in v1])
                    dx1[i0][i1] = np.concatenate([vv.ravel() for vv in dx1[i0][i1]])

        # points
        for i0, v0 in dx0.items():
            for i1, v1 in v0.items():
                if len(v1) > 0:
                    ax0.plot(v1, dx1[i0][i1], '.')

        # concatenate
        x0_all = np.concatenate([
            np.concatenate([v1 for v1 in v0.values() if len(v1) > 0])
            for v0 in dx0.values()
        ])
        x1_all = np.concatenate([
            np.concatenate([v1 for v1 in v0.values() if len(v1) > 0])
            for v0 in dx1.values()
        ])

        # binning
        out = scpstats.binned_statistic_2d(
            x0_all,
            x1_all,
            None,
            statistic='count',
            bins=(cbin0, cbin1),
            expand_binnumbers=True,
        )

        # set binned
        binned = out.statistic

        # plot binning
        im = ax1.imshow(
            binned.T,
            origin='lower',
            interpolation='nearest',
            aspect='equal',
            extent=(cbin0[0], cbin0[-1], cbin1[0], cbin1[-1]),
        )

        plt.colorbar(im, ax=ax1, cax=ax2)
        ax1.set_xlim(cbin0[0], cbin0[-1])
        plt.show()

    import pdb
    pdb.set_trace()     # DB




import warnings


import numpy as np
import datastock as ds


from ..spectro import _rockingcurve


# ###############################################################
# ###############################################################
#                   Rocking curve
# ###############################################################


def rocking_curve(coll=None, key=None):

    # ------------
    # check inputs

    lok = list(coll.dobj.get('crystal', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # --------
    # compute

    raise NotImplementedError
    _rockingcurve.compute_rockingcurve(
    )

    # -------
    # store

    return


# ###############################################################
# ###############################################################
#                   get bragg lambda
# ###############################################################


def _bragglamb(
    coll=None,
    key=None,
    lamb=None,
    bragg=None,
    norder=None,
    rocking_curve=None,
):
    """ Return bragg angle

    If bragg provided, return lamb
    If lamb provided, return bragg

    If lamb is provided return corresponding:
        - bragg angle (simple bragg's law)
        - bragg angle + reflectivity, interpolated on rocking_curve

    """

    # ------------
    # check inputs

    # key
    lok = list(coll.dobj.get('crystal', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # norder
    norder = ds._generic_check._check_var(
        norder, 'norder',
        types=int,
        default=1,
    )

    # rocking_curve
    dmat = coll.dobj['crystal'][key]['dmat']
    lok = [False]
    if dmat is None or dmat.get('drock', {}).get('delta_bragg') is None:
        defrock = False
        lok = (False,)
    else:
        defrock = False
        lok = (False, True)

    rocking_curve = ds._generic_check._check_var(
        rocking_curve, 'rocking_curve',
        types=bool,
        default=defrock,
        allowed=lok,
    )

    # -------------------
    # no input => default

    if bragg is None and lamb is None:
        if dmat is None or dmat.get('target') is None:
            msg = (
                f"Crystal '{key}' has no target lamb!\n"
                f"dmat:\n{dmat}"
            )
            raise Exception(msg)
        else:
            lamb = np.r_[dmat['target']['lamb']]

    if bragg is not None:
        bragg = np.atleast_1d(bragg).astype(float)
    if lamb is not None:
        lamb = np.atleast_1d(lamb).astype(float)

    dist = coll.dobj['crystal'][key]['dmat']['d_hkl']
    if rocking_curve is True:
        delta_bragg = dmat['drock']['delta_bragg']

    # -------------
    # bragg vs lamb

    if bragg is not None and lamb is None:

        if rocking_curve is True:
            lamb = 2. * dist * np.sin(bragg - delta_bragg) / norder

        else:
            lamb = 2. * dist * np.sin(bragg) / norder

    elif lamb is not None and bragg is None:

        if rocking_curve is True:
            bragg = np.arcsin(norder * lamb / (2.*dist)) + delta_bragg

        else:
            bragg = np.arcsin(norder * lamb / (2.*dist))

    else:
        msg = "Interpolate on rocking curve"
        raise NotImplementedError(msg)

    return bragg, lamb


# ###############################################################
# ###############################################################
#                   Ideal configurations
# ###############################################################


def _ideal_configuration_check(
    coll=None,
    key=None,
    configuration=None,
    defocus=None,
    strict_aperture=None,
    # parameters
    cam_on_e0=None,
    # johann-specific
    cam_tangential=None,
    # pinhole-specific
    cam_dimensions=None,
    focal_distance=None,
    # store
    store=None,
    key_cam=None,
    key_aperture=None,
    aperture_dimensions=None,
    pinhole_radius=None,
    cam_pixels_nb=None,
    # returnas
    returnas=None,
):
    # --------------
    # geometry

    # key
    lok = list(coll.dobj.get('crystal', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # gtype
    gtype = coll.dobj['crystal'][key]['dgeom']['type']

    # configuration
    if gtype == 'planar':
        conf = ['pinhole']
    elif gtype == 'cylindrical':
        conf = ['pinhole', 'von hamos', 'johann']
    elif gtype == 'spherical':
        conf = ['pinhole', 'johann']
    elif gtype == 'toroidal':
        conf = []

    if isinstance(configuration, str):
        configuration = configuration.lower()

    configuration = ds._generic_check._check_var(
        configuration, 'configuration',
        types=str,
        allowed=conf,
    )

    # defocus
    defocus = float(ds._generic_check._check_var(
        defocus, 'defocus',
        types=(float, int),
        default=0.,
    ))

    # strict_aperture
    strict_aperture = ds._generic_check._check_var(
        strict_aperture, 'strict_aperture',
        types=bool,
        default=True,
    )

    # cam_on_e0
    cam_on_e0 = ds._generic_check._check_var(
        cam_on_e0, 'cam_on_e0',
        types=bool,
        default=True,
    )

    # cam_tangential
    cam_tangential = ds._generic_check._check_var(
        cam_tangential, 'cam_tangential',
        types=bool,
        default=True,
    )

    # cam_dimensions
    cam_dimensions = ds._generic_check._check_flat1darray(
        cam_dimensions, 'cam_dimensions',
        dtype=float,
        size=[1, 2],
        sign='> 0.',
    )

    if cam_dimensions.size == 1:
        cam_dimensions = cam_dimensions * np.r_[1., 1.]

    # --------------
    # special cases

    # focal_distance
    if configuration == 'pinhole' and gtype == 'planar':
        focal_distance = ds._generic_check._check_var(
            focal_distance, 'focal_distance',
            types=(float, int),
            sign='> 0.',
        )
        focal_distance = float(focal_distance)

    # --------------
    # store-specific

    # store
    store = ds._generic_check._check_var(
        store, 'store',
        types=bool,
        default=False,
    )

    if store is True:

        # key_cam
        lout = list(coll.dobj.get('camera', {}).keys())
        key_cam = ds._generic_check._check_var(
            key_cam, 'key_cam',
            types=str,
            default=f'{key}_cam',
            excluded=lout,
        )

        # key_aperture
        if configuration != 'johann':

            if configuration == 'pinhole':

                lc = [pinhole_radius is None, aperture_dimensions is None]
                if np.sum(lc) != 1:
                    msg = "Provide pinhole_radius xor aperture_dimensions!"
                    raise Exception(msg)

                if pinhole_radius is not None:
                    ap = 'pinhole'
                else:
                    ap = 'slit'
            else:
                ap = 'slit'

            lout = list(coll.dobj.get('aperture', {}).keys())
            key_aperture = ds._generic_check._check_var(
                key_aperture, 'key_aperture',
                types=str,
                default=f'{key}_{ap}',
                excluded=None,
            )

            key_aperture_in = key_aperture in lout

        else:
            key_aperture = None
            key_aperture_in = False

        # cam_pixels_nb
        cam_pixels_nb = ds._generic_check._check_flat1darray(
            cam_pixels_nb, 'cam_pixels_nb',
            dtype=int,
            size=[1, 2],
            sign='> 0.',
        )

        if cam_pixels_nb.size == 1:
            cam_pixels_nb = cam_pixels_nb * np.r_[1, 1]

        # aperture_dimensions
        if key_aperture_in is False:
            if configuration == 'pinhole':
                if pinhole_radius is not None:
                    pinhole_radius = ds._generic_check._check_var(
                        pinhole_radius, 'pinhole_radius',
                        types=float,
                        sign='> 0.',
                    )
                elif aperture_dimensions is not None:
                    aperture_dimensions = ds._generic_check._check_flat1darray(
                        aperture_dimensions, 'aperture_dimensions',
                        dtype=float,
                        size=[1, 2],
                        sign='> 0.',
                    )

            elif configuration == 'von hamos':
                aperture_dimensions = ds._generic_check._check_flat1darray(
                    aperture_dimensions, 'aperture_dimensions',
                    dtype=float,
                    size=[1, 2],
                    sign='> 0.',
                )

            if aperture_dimensions is not None:
                if len(aperture_dimensions) == 1:
                    aperture_dimensions = aperture_dimensions * np.r_[1., 1.]

    # returnas
    returnas = ds._generic_check._check_var(
        returnas, 'returnas',
        default=False if store else dict,
        allowed=[False, dict, list],
    )

    if store is False and returnas is list:
        msg = "returnas = list only of store = True"
        raise Exception(msg)

    return (
        key, gtype, configuration,
        defocus, strict_aperture,
        cam_on_e0, cam_tangential,
        store, key_cam, key_aperture, key_aperture_in,
        cam_pixels_nb, aperture_dimensions,
        focal_distance, pinhole_radius,
        returnas,
    )


def _ideal_configuration(
    coll=None,
    key=None,
    configuration=None,
    lamb=None,
    bragg=None,
    norder=None,
    defocus=None,
    strict_aperture=None,
    # parameters
    cam_on_e0=None,
    # johann-specific
    cam_tangential=None,
    # pinhole-specific
    cam_dimensions=None,
    cam_distance=None,
    focal_distance=None,
    # store
    store=None,
    key_cam=None,
    key_aperture=None,
    aperture_dimensions=None,
    pinhole_radius=None,
    cam_pixels_nb=None,
    # returnas
    returnas=None,
):

    # --------------
    # check inputs

    (
        key, gtype, configuration,
        defocus, strict_aperture,
        cam_on_e0, cam_tangential,
        store, key_cam, key_aperture, key_aperture_in,
        cam_pixels_nb, aperture_dimensions,
        focal_distance, pinhole_radius,
        returnas,
    ) = _ideal_configuration_check(
        coll=coll,
        key=key,
        configuration=configuration,
        defocus=defocus,
        strict_aperture=strict_aperture,
        # parameters
        cam_on_e0=cam_on_e0,
        # johann-specific
        cam_tangential=cam_tangential,
        # pinhole-specific
        cam_dimensions=cam_dimensions,
        focal_distance=focal_distance,
        # store
        store=store,
        key_cam=key_cam,
        key_aperture=key_aperture,
        aperture_dimensions=aperture_dimensions,
        pinhole_radius=pinhole_radius,
        cam_pixels_nb=cam_pixels_nb,
        # returnas
        returnas=returnas,
    )

    # bragg / lamb
    bragg = coll.get_crystal_bragglamb(
        key=key,
        lamb=lamb,
        bragg=bragg,
        norder=norder,
    )[0]
    if bragg.size != 1:
        msg = (
            "Please only provide a single lamb or bragg value!\n"
            f"Provided: {bragg}"
        )
        raise Exception(msg)

    # --------------
    # prepare data

    dgeom = coll.dobj['crystal'][key]['dgeom']
    extenthalf = dgeom['extenthalf']
    curve_r = dgeom['curve_r']
    cent = dgeom['cent']
    nin = dgeom['nin']
    e0 = dgeom['e0']
    e1 = dgeom['e1']

    # radius of curvature
    rc = None
    if gtype == 'spherical':
        rc = curve_r[0]
    elif gtype == 'cylindrical':
        icurv = (~np.isinf(curve_r)).nonzero()[0][0]
        rc = curve_r[icurv]

    # unit vectors
    vect_cam = np.cos(bragg) * e0 + np.sin(bragg) * nin
    vect_los = -np.cos(bragg) * e0 + np.sin(bragg) * nin
    if cam_on_e0 is False:
        vect_cam, vect_los = vect_los, vect_cam

    # ----------------------
    # compute configuration
    # ----------------------

    # --------
    # johann

    if configuration == 'johann':

        if rc < 0:
            msg = (
                f"crystal {key} is convex: Johann not possible!\n"
                f" \t- curve_r = {curve_r}"
            )
            raise Exception(msg)

        med = rc * np.sin(bragg)
        sag = -med / np.cos(2.*bragg)

        meridional = cent + med * vect_los
        sagittal = cent + sag * vect_los

        cam_cent = cent + (med + defocus) * vect_cam

        if cam_tangential is True:
            cam_nin = (cent + nin*rc/2. - cam_cent)
            cam_nin = cam_nin / np.linalg.norm(cam_nin)
        else:
            cam_nin = -vect_los

        dout = {
            'meridional': {
                'cent': meridional,
                'dist': med,
            },
            'sagittal': {
                'cent': sagittal,
                'dist': sag,
            },
        }

    # --------
    # von hamos

    elif configuration == 'von hamos':

        if rc < 0:
            msg = (
                f"crystal {key} is convex: von hamos not possible!\n"
                f" \t- curve_r = {curve_r}"
            )
            raise Exception(msg)

        dist_pin = rc / np.sin(bragg)
        pin_cent = cent + dist_pin * vect_los
        pin_nin = vect_los

        cam_cent = cent + (dist_pin + defocus) * vect_cam
        if cam_tangential is True:
            cam_nin = -nin
        else:
            cam_nin = -vect_cam

        dout = {
            'aperture': {
                'cent': pin_cent,
                'nin': pin_nin,
            },
        }

    # --------
    # pinhole

    elif configuration == 'pinhole':

        # pinhole
        if focal_distance is None:
            if gtype == 'cylindrical':
                focus = np.abs(rc) / np.sin(bragg)
            elif gtype == 'spherical':
                focus = np.abs(rc) * np.sin(bragg)
            else:
                msg = "Please provide focal_distance!"
                raise Exception(msg)

        else:
            focus = focal_distance

        pin_cent = cent + focus * vect_los
        pin_nin = vect_los

        # camera
        if cam_distance is None:
            cam_height = cam_dimensions[1]

            if gtype == 'planar' or (gtype == 'cylindrical' and icurv == 0):
                cryst_height = 2. * extenthalf[1]

                if cam_height <= cryst_height:
                    msg = (
                        f"Height for ideal camera of '{key}' too small:\n"
                        f"\t- crystal height (flat): {cryst_height}\n"
                        f"\t- camera height: {cam_height}\n"
                    )
                    raise Exception(msg)

                cam_dist = focal_distance * (cam_height / cryst_height - 1.)

            else:
                if gtype == 'cylindrical':
                    cryst_height = 2. * extenthalf[icurv] * np.abs(rc)

                elif gtype == 'spherical':
                    cryst_height = 2. * extenthalf[1] * np.abs(rc)

                if cam_height >= cryst_height:
                    msg = (
                        f"Height for ideal camera of '{key}' too large:\n"
                        f"\t- crystal height (gtype): {cryst_height}\n"
                        f"\t- camera height: {cam_height}\n"
                    )
                    raise Exception(msg)

                cam_dist = np.abs(rc) * (1. - cam_height / cryst_height)

        else:
            cam_dist = cam_distance

        cam_nin = -vect_cam
        cam_cent = cent + (cam_dist + defocus) * vect_cam

        dout = {
            'aperture': {
                'cent': pin_cent,
                'nin': pin_nin,
            },
        }

    # ----------------------------------
    # complete with missing unit vectors

    if 'aperture' in dout.keys():

        # check against existing aperture is any
        if key_aperture_in is True:

            dd = coll.dobj['aperture'][key_aperture]['dgeom']
            temp_dist = np.linalg.norm(dd['cent'] - pin_cent)
            temp_ang = np.arctan2(
                np.linalg.norm(np.cross(dd['nin'], pin_nin)),
                np.sum(dd['nin'] * pin_nin),
            )

            if not (temp_dist < 1e-6 and np.abs(temp_ang) < 0.01*np.pi/180.):
                # dist = np.linalg.norm(dd['cent'] - pin_cent)
                msg = (
                    f"Ideal configuration '{configuration}' for crystal '{key}':\n"
                    f"Predefined aperture {key_aperture} does not seem fit:\n"
                    f"\t- cent: {temp_dist} m\n"
                    f"\t\t- {key_aperture}: {dd['cent']}\n"
                    f"\t\t- ideal: {pin_cent}\n"
                    f"\t- nin: {temp_ang} deg.\n"
                    f"\t\t- {key_aperture}: {dd['nin']}\n"
                    f"\t\t- ideal: {pin_nin}\n"
                )
                if strict_aperture is True:
                    raise Exception(msg)
                else:
                    warnings.warn(msg)
            del dout['aperture']

        # new aperture
        else:
            ap_e0 = np.cross(e1, dout['aperture']['nin'])
            ap_e0 = ap_e0 / np.linalg.norm(ap_e0)
            ap_e1 = np.cross(dout['aperture']['nin'], ap_e0)
            dout['aperture']['e0'] = ap_e0
            dout['aperture']['e1'] = ap_e1

    cam_e0 = np.cross(e1, cam_nin)
    cam_e0 = cam_e0 / np.linalg.norm(cam_e0)
    cam_e1 = np.cross(cam_nin, cam_e0)

    # ---------
    # return

    dout.update({
        'camera': {
            'cent': cam_cent,
            'nin': cam_nin,
            'e0': cam_e0,
            'e1': cam_e1,
        },
    })

    # ---------
    # store

    if store is True:

        dout = _ideal_configuration_store(
            coll=coll,
            configuration=configuration,
            dout=dout,
            key_cam=key_cam,
            key_aperture=key_aperture,
            cam_dimensions=cam_dimensions,
            cam_pixels_nb=cam_pixels_nb,
            aperture_dimensions=aperture_dimensions,
            pinhole_radius=pinhole_radius,
        )

    # ---------
    # return

    if returnas is dict:
        return dout

    elif returnas is list:
        loptics = [key_cam, key]
        if key_aperture is not None:
            loptics.append(key_aperture)
        return loptics


def _ideal_configuration_store(
    coll=None,
    configuration=None,
    dout=None,
    # store
    key_cam=None,
    key_aperture=None,
    cam_dimensions=None,
    cam_pixels_nb=None,
    aperture_dimensions=None,
    pinhole_radius=None,
):

    # -------
    # camera

    # pixels dimensions, outline and cents
    dim0, dim1 = cam_dimensions
    nx0, nx1 = cam_pixels_nb

    dx0 = dim0 / nx0
    dx1 = dim1 / nx1

    outline_x0 = 0.5 * dx0 * np.r_[-1., 1., 1., -1.]
    outline_x1 = 0.5 * dx1 * np.r_[-1., -1., 1., 1.]

    cents_x0 = 0.5 * dim0 * np.linspace(-1., 1., nx0 + 1)
    cents_x1 = 0.5 * dim1 * np.linspace(-1., 1., nx1 + 1)
    cents_x0 = 0.5 * (cents_x0[:-1] + cents_x0[1:])
    cents_x1 = 0.5 * (cents_x1[:-1] + cents_x1[1:])

    # complement camera
    dout['camera'].update({
        'outline_x0': outline_x0,
        'outline_x1': outline_x1,
        'cents_x0': cents_x0,
        'cents_x1': cents_x1,
    })

    # add camera
    coll.add_camera_2d(
        key=key_cam,
        dgeom=dout['camera'],
    )

    # -------
    # aperture

    if 'aperture' in dout.keys():

        if pinhole_radius is not None:
            theta = np.pi * np.linspace(-1, 1, 50)[:-1]
            outline_x0 = pinhole_radius * np.cos(theta)
            outline_x1 = pinhole_radius * np.sin(theta)
        else:
            # outline
            lx0, lx1 = aperture_dimensions
            outline_x0 = 0.5 * lx0 * np.r_[-1., 1., 1., -1.]
            outline_x1 = 0.5 * lx1 * np.r_[-1., -1., 1., 1.]

        # complement dict
        dout['aperture'].update({
            'outline_x0': outline_x0,
            'outline_x1': outline_x1,
        })

        # add aperture
        coll.add_aperture(
            key=key_aperture,
            **dout['aperture'],
        )

    return dout